"""
Contain the class the represents the game engine, an object 
which runs the game loop.
"""

from play_mode import PlayMode
from game_over_mode import GameOverMode
from full_row_mode import FullRowMode
from title_mode import TitleMode

import pygame
import sys

class Engine:
    """
    The game engine; it runs the game loop.
    
    An C{Engine} has four modes.
        1. B{Play Mode} is the mode in which the user
        play the game.
        2. B{Full Row Mode} is the mode in which the 
        game displays the rows that are full before
        removing it from the board.
        3. B{Game Over Mode} is the mode in which
        the Game Over screen is shown. The user
        has the options to (1) play the game again,
        (2) return to the title, or (3) quit the
        application. 
        4. B{Title Mode} is the mode in which the
        title screen is shown.
        
    @ivar screen: the surface the represents the area
        of the game window.
    @type screen: C{pygame.Surface}
    
    @ivar vera64: 64 point Bitstream Vera font
    @type vera64: C{pygame.font.Font}
    @ivar vera32: 32 point Bitstream Vera font
    @type vera32: C{pygame.font.Font}
    
    @ivar play_mode: the Play Mode
    @type play_mode: C{play_mode.PlayMode}
    @ivar game_over_mode: the Game Over Mode
    @type game_over_mode: C{game_over_mode.GameOverMode}
    @ivar full_row_mode: the Full Row Mode
    @type full_row_mode: C{full_row_mode.FullRowMode}
    @ivar title_mode: the Title Mode
    @type title_mode: C{title_mode.TitleMode}
    
    @ivar current_mode: the current mode the engine is in
    @type current_mode: C{mode.Mode}
    
    @ivar clock: Pygame's clock
    @type clock: C{pygame.time.Clock}
    
    @ivar max_frame_rate: the maximum frame rate of the game,
        defaulted to 100
    @type max_frame_rate: number
    """
    def init(self):
        """
        Create and intialize pygame and the engine itself.
        """
        # Initialize pygame.
        pygame.init()        
        # Create the screen surface.
        self.screen = pygame.display.set_mode((640, 480), 0, 32)
        # Set the caption of the window.
        pygame.display.set_caption("Tetris")
        
        # Load Bitstream Vera monospace font.
        self.vera64 = pygame.font.Font("VeraMono.ttf", 64)
        self.vera32 = pygame.font.Font("VeraMono.ttf", 32)

        # Create the four modes.
        self.play_mode = PlayMode(self)
        self.game_over_mode = GameOverMode(self)
        self.full_row_mode = FullRowMode(self)
        self.title_mode = TitleMode(self)        
            
        # Set the current mode to the Title Mode.
        self.current_mode = self.title_mode
        self.change_mode(self.title_mode)
        
        # Create the clock.
        self.clock = pygame.time.Clock()
        
        # Initialize the maximum frame rate constant.
        self.max_frame_rate = 100
        
    def change_mode(self, mode):
        """
        Change the engine's mode to the specified mode.
        """
        old_mode = self.current_mode
        self.current_mode = mode
        self.current_mode.switched_into(old_mode)
        
    def handle(self, event):
        """
        Catches the C{QUIT} event and
        closes the program if one is caught.
        
        This method is meant to be called by
        instances of C{Mode}.
        """
        if event.type == pygame.locals.QUIT:
            sys.exit()
                
    def game_loop(self):
        """
        Run the game loop.
        """
        while True:
            # Limit the frame rate to 100 frames/s
            self.clock.tick(self.max_frame_rate)
            
            # Let the current mode initialize itself
            # before handling events.
            self.current_mode.preprocess()            
            
            # Let the current mode handle events.
            for event in pygame.event.get():
                self.current_mode.handle(event)
            
            # Let the current mode update its states.
            self.current_mode.process()
            
            # Let the current mode render the screen.
            self.current_mode.render(self.screen)
            # Then update it.
            pygame.display.update()