"""
Contains the class FullRowMode
"""

from mode import *
from local_time_marker import *

import pygame
import sys

WHITE = 0
NORMAL = 1

class FullRowMode(Mode, LocalTimeMarker):
    """
    A mode that makes the rows that are full blink
    to times before returning to the Play Mode.
    
    @ivar full_rows: the list of indices of full rows
    @type full_rows: list of integers
    
    @ivar blink_delay: the number of milliseconds
        from when the full rows start blinking
        (turn white) to when the full rows stop
        blinking (turn back to their normal colors).
        This is also the time from when the rows
        turn to their normal colors to when the 
        rows become white again. It is defaulted to
        200 milliseconds.
    @type blink_delay: integer
    
    @ivar display_state: an enumerated value
        (represented by an integer) that tells whether
        the blocks in the full rows should be white or 
        should be in their original colors at the moment;
        possible values are C{WHITE} (0) and C{NORMAL} (1)
    @type display_state: integer
    """
    def __init__(self, engine):
        """
        Initialize the mode.
        """
        Mode.__init__(self, engine)
        LocalTimeMarker.__init__(self)
        
        self.full_rows = []        
        self.blink_delay = 200
        self.display_state = WHITE
        
    def render(self, surface):
        """
        Render the game screen to the given surface.
        """
        # First we render the scene using Play Mode's
        # render method so that we get the tetris board
        # and all the labels.
        self.engine.play_mode.render(surface)
        
        board_surface = self.engine.play_mode.get_board_surface(surface)
        # Then, if blocks in the full rows should be
        # white now...
        if self.display_state == WHITE:
            # We render the blocks with the color white
            # one by one.
            for row in self.full_rows:
                for col in range(self.engine.play_mode.board_width):
                    self.engine.play_mode.render_block(board_surface, (col, row), (255,255,255))
    
    def handle(self, event):
        """
        Call the Play Mode's C{handle} method so
        that you can hold the down button
        and have the tetromino move down immediately
        after the Full Row Mode changes back
        to the Play Mode.
        """
        self.engine.play_mode.handle(event)
    
    def preprocess(self):
        """
        Call the Play Mode's C{preprocess} method.
        """
        self.engine.play_mode.preprocess()
    
    def process(self):
        """
        Change C{display_state} to the appropriate
        value based on the time elapsed. Also,
        change the mode back to Play Mode if 
        the full rows have blinked two times.
        """
        self.update_local_time() 
        
        # If it has been more than blink_delay
        # change display_state.
        if self.millisecs_since_local_mark("blink") > self.blink_delay:
            self.mark_local_time("blink")
            if self.display_state == WHITE:
                self.display_state = NORMAL
            else:
                self.display_state = WHITE
                self.counter += 1
            # Switch back to Play Mode if
            # the full rows have blinked two times.
            if self.counter >= 2:
                self.leave_local_time()
                self.engine.change_mode(self.engine.play_mode)
        
    def switched_into(self, from_mode):
        """
        Prepare the mode for display 
        blinking full rows if the mode
        is switched into from the Play Mode.
        """
        self.enter_local_time()
        if from_mode == self.engine.play_mode:
            self.mark_local_time("blink")
            self.display_state = WHITE
            self.counter = 0