"""
Contain the class KeySensor.
"""

import pygame

from time_marker import TimeMarker

class KeySensor(TimeMarker):
    """
    Mixin class that contains the capability to sense keyboard events at
    a high level. In particular, it has the ability to sense:
        - whether a certain keyboard button was pressed in the current frame,
        - whether a certain keyboard button was released in the current frame,
        - whether a certain keyboard button is being held in the current frame, and
        - whether a certain keyboard button is "typed" in the current frame.
    
    What does it mean for a keyboard button to be "typed?"
    ======================================================
        In an editor, when you press a keyboard button, say the button 'a',
        and hold it. The following things happen:
            1. As soon as you press the button, the letter 'a' shows
                up on your program.
            2. As you hold the keyboard, the next 'a' does not show
                up for a relatively long period of time, say 500ms.
            3. After the second 'a' appeared, subsequence 'a's will show
                up with higher frequency, say one 'a' per 150 ms.
            4. When you release the button, no more 'a' will show up.            
        We say that a keyboard button is I{typed} in a certain frame when
        the character corresponding to that button shows up in that frame.
        
        Formally, "being typed in this frame" is equivalent to the
        following conditions:
            
            - The keyboard button was just pressed in this frame OR            
            - The following three conditions hold simultaneously in that frame:
                 1. The user is currently pressing the button.
                 2. The button has been held for longer than a certain fixed amount of time.
                 We call this amount of time the I{initial delay}.
                 3. Since the last moment the same keyboard button was typed,
                 it has been longer than another certain fixed amount of time.
                 We call this amount of time the I{repeat delay}.
                 
        Typically, the initial delay is 500ms, and the repeat delay is 100ms to 150ms.
        You can set the initial delay and repeat delay to your liking by setting 
        the C{key_initial_delay} and C{key_repeat_delay} attributes, respectively.
        
    Using KeySensor
    ===============        
        KeySensor is meant to use as a mixin that adds keyboard event handling capibility
        to subclasses of C{mode.Mode}. In the C{process} method of such a subclass,
        the subclass can check whether a keyboard button is typed in that frame by calling::
            
            self.is_key_typed(key)
        
        where C{key} is a Pygame constant in C{pygame.locals} that represents the keyboard
        button. For example, the following code snippet checks whether the button 'a' was 
        typed, and, if so, it prints a message::
        
            if self.is_key_typed(K_a):
                print "Button 'a' was typed."
        
        KeySensor also comes with methods C{is_key_down}, C{is_key_up}, and C{is_key_pressed},
        which can be used to check with a certain keyboard button was just pressed in that
        frame, was just released in that frame, or is still being pressed in that frame, 
        respectively.
        
        A number contracts have be followed so that the above four functions produce
        correct outputs.
            1. In the C{preprocess} method, you have to call C{self.clear_key_flags()}.
            2. In the C{handle} method, you have to call C{self.handle_key_events(event}.
            3. In the C{process} method, you have to call C{self.process_key_events()} before
            calling any of the above four functions.
        Please see C{menu_mode.py} for an example of how to following the above contracts.
        
    @ivar key_initial_delay: the initial delay in millisecond
    @type key_initial_delay: number
    
    @ivar key_repeat_delay: the repeat delay in millisecond
    @type key_repeat_delay: number
    """
    def __init__(self, initial_delay=500, repeat_delay=150):
        """
        Initialize the internals of KeySensor, and set
        the initial delay and the repeat delay to the
        specified values.
        """        
        # Call the superclass's constructor.
        TimeMarker.__init__(self)
        
        ######################################        
        # Create four dictionaries that maps 
        # Pygame's code for keyboard buttons 
        # to boolean values.                                 
        ######################################
        
        self.__key_down = {}         
        # __key_down[key] = True if and only if there is 
        #    a KEYDOWN event with that key in the current 
        #    frame.
        
        self.__key_up = {}
        # __key_up[key] = True if and only if there is 
        #    a KEYUP event with that key in the current 
        #    frame.
        
        self.__key_pressed = {}
        # __key_pressed[key] = True if and only if the
        #    key is being pressed (i.e., has not been
        #    released) in the current frame.
         
        self.__key_typed = {}
        # __key_typed[key] = True if and only if the
        #    key is being typed in the current 
        #    frame.
                
        ######################################
        # To prevent bugs, I wrote the code 
        # so that there can be only one key
        # being pressed, up, or down in a 
        # frame. That is, when either a new
        # KEYUP or KEYDOWN event shows up,
        # I reset all of these dictionaries.
        #
        # Please see the method
        # handle_key_event below. 
        ######################################
        
        # Set the initial delay and the repeat delay.
        self.key_initial_delay = initial_delay
        self.key_repeat_delay = repeat_delay
        
    def is_key_pressed(self, key):
        """
        Return True if the given key is pressed
        in the current frame; return False otherwise.
        
        @rtype: boolean
        """
        if key in self.__key_pressed:
            return self.__key_pressed[key]
        else:
            return False
        
    def is_key_typed(self, key):
        """
        Return True if the given key is typed
        in the current frame; return False otherwise.
        
        @rtype: boolean
        """
        if key in self.__key_typed:
            return self.__key_typed[key]
        else:
            return False
    
    def is_key_down(self, key):
        """
        Return True if the given key was just pressed
        in the current frame; return False otherwise.
        
        @rtype: boolean
        """
        if key in self.__key_down:
            return self.__key_down[key]
        else:
            return False
        
    def is_key_up(self, key):
        """
        Return True if the given key was just released
        in the current frame; return False otherwise.
        
        @rtype: boolean
        """
        if key in self.__key_up:
            return self.__key_up[key]
        else:
            return False
        
    def __reset_all_key_dictionaries(self):
        """
        A convenience method that resets
        all internal dictionaries.
        """
        for key in self.__key_down:
            self.__key_down[key] = False
        for key in self.__key_pressed:
            self.__key_pressed[key] = False
        for key in self.__key_typed:
            self.__key_typed[key] = False
        for key in self.__key_up:
            self.__key_up[key] = False
            
    def handle_key_events(self, event):
        """
        Handle a keyboard event. This method should
        be called in the C{handle} method of 
        C{mode.Mode} subclasses.
        """
        if event.type == pygame.locals.KEYDOWN:
            # As stated earlier, we reset 
            # all the dictionaries when a new
            # KEYDOWN event shows up.
            self.__reset_all_key_dictionaries()
            
            # Set the flags appropriately.
            self.__key_pressed[event.key] = True
            self.__key_down[event.key] = True
            
            # Mark the time the key is pressed.
            # This will be used to figure out
            # how long has the used been pressing
            # the key.
            self.mark_time(str(event.key) + "down")
                                    
        elif event.type == pygame.locals.KEYUP:
            # Just like when we handle
            # a KEYDOWN event, we reset
            # the dictionaries whenever a
            # KEYUP event shows up as well.
            self.__reset_all_key_dictionaries()
            
            # Then we set the flag appropriately.
            self.__key_up[event.key] = True
    
    def clear_key_flags(self):
        """
        Initialize internal flags of 
        KeySensor before handling inputs.
        This method should be called in the
        C{preprocess} method of C{mode.Mode}
        subclasses.
        """        
        # Since the values of __key_up,
        # __key_down, and __key_typed changes
        # on a per-frame basis. We reset
        # the three dictionaries before
        # handling events every frame.
        for key in self.__key_up:
            self.__key_up[key] = False
        for key in self.__key_down:
            self.__key_down[key] = False
        for key in self.__key_typed:
            self.__key_typed[key] = False
               
    def process_key_events(self):
        """
        Process keyboard events (which
        have already been handled by
        C{handle_key_events}) in order to
        decide which key is typed in 
        the current frame. This method
        should be called in the C{process}
        method of C{mode.Mode} subclasses
        before using C{is_key_typed}.
        """
        for key in self.__key_pressed:
            # We loop through all the keys 
            # that are currently being pressed.
            if self.is_key_pressed(key):
                # The variable "hold" is True if
                # the user has been pressing the key
                # for longer than key_initial_delay                
                hold = False
                if self.millisecs_since_mark(str(key)+"down") > self.key_initial_delay:
                    hold = True
                
                # We check whether the condition for
                # the key to be typed as stated above
                # is true.
                if (self.is_key_down(key) or 
                    hold and self.millisecs_since_mark(str(key)+"typed") > self.key_repeat_delay):
                    # If so, we mark that the key is 
                    # being typed now.
                    self.mark_time(str(key)+"typed")
                    # And the physically indicate that
                    # the key is typed.
                    self.__key_typed[key] = True