"""
Contains the mixin class LocalTimeMarker
"""

import pygame
from time_marker import TimeMarker

class LocalTimeMarker(TimeMarker):
    """
    Mixin class that adds the capability to mark 
    a certain moment in time and figure out how 
    many milliseconds the game spent in a specific mode
    has elapsed since the marked time.
    
    The class TimeMarker reports the time elapsed based
    on a "global" time which flows regardless of which 
    mode the game is in. However, LocalTimeMarker's time 
    only flows in the mode that inherits it, and stops
    outside the mode.
    
    If you want a mode to LocalTimeMarker's functionality,
    you needs to follow the following contracts:
        1. When the engine changes into the mode from
           another mode, C{enter_local_time} must be called.
        2. When the engine is about to change from the
           mode to another mode, C{leave_local_time} must
           be called.
        3. C{update_local_time} has to be called once per frame.
    Then, you can call::
        
        self.mark_local_time("name-of-the-event")
        
    to associate an event with the local time at the moment, and
    then call::
        
        self.millisecs_since_local_mark("name-of-the-event")
        
    to get the number of milliseconds from the local time the
    event was last marked.
    
    You can see sample usage of LocalTimeMarker in 
    C{full_row_mode.py}.
    
    @ivar __local_time: the local time since the instance of
        LocalTimeMarker was in milliseconds
    @type local_time: integer
    """
    def __init__(self):
        """
        Initialize the local time marker.
        """
        TimeMarker.__init__(self)
        
        self.mark_time("__last_local_time__")
        self.__local_time  = 0
        
        self.__marked_local_times = {}
    
    def mark_local_time(self, event_name):
        """
        Associate C{event_name} with the current local time.
        
        @param event_name: a value used to denote
            the event to be marked.
        @type event_name: hashable
        """
        self.__marked_local_times[event_name] = self.__local_time
    
    def millisecs_since_local_mark(self, event_name):
        """
        Return the number of milliseconds in local time 
        since the event designated by C{event_name}
        was marked.
        
        @param event_name: a value used to denote
            the event to be marked.
        @type event_name: hashable
        
        @rtype: integer
        """
        return self.__local_time - self.__marked_local_times[event_name]
    
    def update_local_time(self):
        """
        Call this method every frame to advance the local time.
        """
        self.__local_time += self.millisecs_since_mark("__last_local_time__")
        self.mark_time("__last_local_time__")
        
    def leave_local_time(self):
        """
        Call this method to stop the local time; in other words,
        every time the engine changes to another mode.
        """
        self.update_local_time()
    
    def enter_local_time(self):
        """
        Call this method to start the local time again; 
        in other words, every time the engine switches to
        this mode.
        """
        self.mark_time("__last_local_time__")