"""
Contain the class MenuMode.
"""

from mode import *
from time_marker import *
from key_sensor import *

import pygame
from exceptions import *

class MenuMode(Mode, KeySensor):
    """
    Abstract class representing a mode of the game 
    in which the user is presented with a menu.
    The menu has one big "header" and a number of smaller
    choices. The user can switch between the choices
    by the Up and Down keys. Then, he/she can select
    a particular choices by the Enter key.
    """
    def __init__(self, engine, header, choices, header_font, choice_font):
        """
        Create a Menu Mode with the given header and choices.
        
        Example usage::
        
            class MyMenuMode(MenuMode)
                def __init__(self, engine)
                    MenuMode.__init__(self, "My Menu",
                                      ["Choice #1", "Choice #2"],
                                      engine.vera64, engine.vera32)
        
        @param engine: the game engine
        @type engine: C{engine.Engine}
        
        @param header: the text of the header
        @type header: string
        
        @param choices: the list of choices, each of
            which is a string.
        @type choices: list of string
        
        @param header_font: the font used to render
            the header.
        @type header_font: C{pygame.font.Font}
        
        @param choice_font: the font used to render
            the choices
        @type choice_font: C{pygame.font.Font} 
        """
        # Call the superclasses' constructors.
        Mode.__init__(self, engine)        
        KeySensor.__init__(self, 500, 150)
        
        # Store the fields
        self.header = header
        self.choices = choices
        self.header_font = header_font
        self.choice_font = choice_font
        
        # Set the "current choice" to be the first choice.
        self.current_choice = 0
        
        # Create surface of the header.
        self.header_surface = self.header_font.render(
            self.header, True, (255, 255, 255))
        
        # Create surface for the choices.
        self.choice_unselected_surfaces = []    
        self.choice_selected_surfaces = []
        for choice in self.choices:
            self.choice_unselected_surfaces.append(
                self.choice_font.render(choice, True, (255,255,255)))
            self.choice_selected_surfaces.append(
                self.choice_font.render('['+choice+']', True, (255,255,255)))
        
        # Compute the line size of the header and the choices.
        self.header_linesize = self.header_font.get_linesize()
        self.choice_linesize = self.choice_font.get_linesize()
                                                             
        # Compute the vertical span of the header and the choices
        self.menu_height = 2 * self.header_linesize
        self.menu_height += (self.choice_linesize * len(self.choices) 
            + self.choice_linesize / 2 * (len(self.choices)-1))        
        
    def switched_into(self, from_mode):
        """
        Set the current choice to the first choice, 
        """
        self.current_choice = 0        
        
    def handle(self,event):
        """
        Call the Engine's handle method and
        KeySensor's handle method.
        """
        self.engine.handle(event)
        self.handle_key_events(event)        
    
    def preprocess(self):
        """
        Call KeySensor's clear_key_flags.
        """
        self.clear_key_flags()
        
    def process(self):
        """
        Change the current choice based on
        the input.
        """
        # Let KeySensor determine if semething
        # is typed.
        self.process_key_events()
        
        # If the down arrow is typed, change the choice
        # to the one below.
        if self.is_key_typed(pygame.locals.K_DOWN):
            self.current_choice = (self.current_choice + 1) % len(self.choices)
                
        # If the up arrow is typed, change the choice
        # to the one above.
        if self.is_key_typed(pygame.locals.K_UP):                
            self.current_choice = (self.current_choice - 1) % len(self.choices)
        
        # Call choice_selected if the user presses Return.
        if self.is_key_down(pygame.locals.K_RETURN):
            self.choice_selected(self.current_choice)
    
    def render(self, surface):
        """
        Fill the given surface with a black background
        and then render the menu's text on top of it.
        """
        surface.fill((0,0,0))
        self.render_menu(surface)
    
    def render_menu(self, surface):
        """
        Render only the menu's text to the given surface. 
        """
        surface_width, surface_height = surface.get_size()
        # Calculate (x,y) coordinate so that the header
        # is centered horizontally.
        y = (surface_height - self.menu_height)/2
        x = (surface_width - self.header_surface.get_width())/2
        surface.blit(self.header_surface, (x,y))
        y += 2*self.header_linesize
        
        # Blit the choices from top to bottom.
        for i in range(len(self.choices)):
            if i != self.current_choice:
                # If the choice is not equal to the current
                # choice, use the surface for the 
                # unselected case.
                to_blit = self.choice_unselected_surfaces[i]
            else:
                # Otherwise, use the surface for the 
                # selected case.
                to_blit = self.choice_selected_surfaces[i]
            # Compute the horizontal offset.
            x = (surface_width - to_blit.get_width())/2
            # Then blit the surface determined above.
            surface.blit(to_blit, (x,y))
            # Then move down the screen.
            y += self.choice_linesize + self.choice_linesize/2
    
    def choice_selected(self, choice_index):
        """
        Subclasses of menu mode should override this
        method to implement behaviors when a choice is
        selected.
        
        @param choice_index: the index of the selected
            choice, starting from 0.
        @type choice_index: integer
        """
        raise NotImplementedError()