"""
Contains the PlayMode class.

@var tetromino_colors: A dictionary that maps tetromino 
    shapes represented by a character to its color.
"""
from mode import Mode
from tetris_logic import *
from label import *
from key_sensor import *

import pygame
import sys
import random
from pygame.locals import Rect

tetromino_colors = {'i': (255,0,0),
                    'j': (255,255,0),
                    'l': (255,0,255),
                    'o': (0,0,255),
                    's': (0,255,255),
                    't': (0,255,0),
                    'z': (255,128,0)}

class PlayMode(Mode, KeySensor):
    """
    The mode that handles the game play.
    
    @ivar board: the tetris board.
        Its size is defaulted to 10x18
    @type board: C{tetris_logic.TetrisBoard}
    
    @ivar block_size: the side length of a block in pixels, 
        defaulted to 25.
    @type block_size: integer
    
    @ivar player_piece: the tetromino that
        the player is controlling
    @type player_piece: C{tetris_logic.Tetromino}
    
    @ivar next_piece: the next tetromino that
        the player will be controlling.
    @type next_piece: C{tetris_logic.Tetromino}
    
    @ivar descend_delay: the number of milliseconds
        until the tetromino piece the player controls
        moves down one row automaticallly.
    @type descend_delay: C{tetris_logic.Tetromino}
    
    @ivar full_rows_created: the number of full rows
        created by the player
    @type full_rows_created: integer
    
    @ivar score: the score
    @type score: integer
    
    @ivar next_label: the label showing the word "next"
    @type next_label: C{label.Label}
    
    @ivar score_text_label: the label showing the word "score"
    @type score_text_label: C{label.Label}

    @ivar score_num_label: the label showing the numerical score
    @type score_num_label: C{label.Label}
    
    @ivar rows_text_label: the label showing the word "rows"
    @type rows_text_label: C{label.Label}
    
    @ivar rows_num_label: the label showing the number
        of full rows the player has created so far.
    @type rows_num_label: C{label.Label}
    
    @ivar labels: a list containing the five UI labels
    @type labels: list of C{label.Label}
    
    @ivar board_rect: a (x,y,w,h)-tuple representing
        the rectangle in screen coordinate system that
        bounds the tetris board.
    @type board_rect: 4-tuple of integers
    
    @ivar next_piece_area_rect: a (x,y,w,h)-tuple representing
        to rectangle in screen coordinate system that
        bounds the area showing the next tetris piece.
    @type next_piece_area_rect: 4-tuple of integers
    """
    def __init__(self, engine):
        """
        Create the PlayMode and initialize
        its instance variables.
        
        @param engine: the game engine
        @type engine: C{engine.Engine}
        """
        # Call superclasses' constructors
        Mode.__init__(self, engine)
        KeySensor.__init__(self, 200, 50)
        
        # ======================================
        # = Initialize the instance variables. =
        # ======================================
        self.board_width = 10
        self.board_height = 18
        self.board = TetrisBoard(self.board_width, self.board_height)
        
        self.block_size = 25
        
        self.player_piece = None
        self.next_piece = None
        
        self.descend_delay = 1000        
        
        self.full_row_created = 0
        self.score = 0
            
        # ===============================
        # = Create the UI labels and    =
        # = rectangles demarcating the  =
        # = game board and the area     =
        # = showing the next tetromino. =
        # ===============================
        self.console_width = 430
        self.console_height = 450
        
        self.console_x = (self.engine.screen.get_size()[0] - 
                          self.console_width)/2
        self.console_y = (self.engine.screen.get_size()[1] - 
                          self.console_height)/2
        
        self.board_rect = Rect(self.console_x, 
                               self.console_y, 
                               self.board_width * self.block_size,
                               self.board_height * self.block_size)
        
        self.infocol_x = self.console_x + self.block_size * self.board_width + 30
        
        self.next_label = Label("next", 
                                self.engine.vera32, 
                                (self.infocol_x, self.console_y),
                                (255, 255, 255))
                         
        self.next_piece_area_rect = Rect(self.infocol_x, 
                                         self.next_label.position[1] + 40,
                                         6 * self.block_size,
                                         4 * self.block_size)
        
        self.score_text_label = Label("score",
                                      self.engine.vera32,
                                      (self.infocol_x, 
                                       self.next_piece_area_rect[1] + 
                                       self.next_piece_area_rect[3] + 40),
                                      (255,255,255))
        
        self.score_num_label = Label("000000",
                                     self.engine.vera32,
                                     (self.infocol_x, 
                                      self.score_text_label.position[1] + 40),
                                     (255,255,255))

        self.rows_text_label = Label("rows",
                                     self.engine.vera32,
                                     (self.infocol_x, 
                                      self.score_text_label.position[1] + 140),
                                     (255,255,255))
        
        self.rows_num_label = Label("000000",
                                    self.engine.vera32,
                                    (self.infocol_x, 
                                     self.rows_text_label.position[1] + 40),
                                    (255,255,255))
        
        self.labels = [self.next_label,
                       self.score_text_label, self.score_num_label,
                       self.rows_text_label, self.rows_num_label]
            
    def new_board(self):
        """
        Empty tetris board and create
        two initial tetrominoes.
        """
        # Mark the time of last descent
        # to be the time of creation,
        # so that the first tetromino
        # does not go down immediately
        # when the game start.
        self.mark_time("last descent")
        
        self.board.clear()
        self.full_rows_created = 0
        self.score = 0
        
        # Create two new tetrominoes
        # so that both player_piece and
        # next_piece are well defined.
        self.prepare_next_piece()
        self.prepare_next_piece()    
                
    def get_board_surface(self, surface):
        """
        Return a subsurface of the given surface
        whose area is the rectangle bounding
        the tetris board.
        
        @param surface: a surface
        @type surface: C{pygame.Surface}
        
        @rtype: C{pygame.Surface}
        """
        return surface.subsurface(self.board_rect)
        
    def get_next_piece_area_surface(self, surface):
        """
        Return a subsurface of the given surface
        whose area is the area used to show
        the next tetris piece.
        
        @param surface: a surface
        @type surface: C{pygame.Surface}
        
        @rtype: C{pygame.Surface}
        """
        return surface.subsurface(self.next_piece_area_rect)
    
    def render(self, surface):
        """
        Make the background grey and render
        all the UI elements to the given
        surface.
        
        @param surface: a Pygame surface
        @type surface: C{pygame.Surface}    
        """
        # Make the background grey.
        surface.fill((128,128,128))
        
        # Render the tetris board.
        self.render_board(self.get_board_surface(surface))
        
        def num_string_with_leading_zeros(num, positions):
            """
            Convert the given integer to a string representing
            the same integer with length at least C{position} by
            padding zeros to the front of the number if the
            number is too short.
            
            @param num: the integer to be turned into a string.
            @type num: integer
            
            @param positions: the number of characters in the resulting
                string.
            @type positions: integer
            
            @rtype: string
            """
            s = str(num)
            if len(s) < positions:
                return "0"*(positions-len(s)) + s
            else:
                return s
            
        # Compute the text representing the score 
        # and the number of rows created by the player.
        self.score_num_label.set_text(num_string_with_leading_zeros(self.score, 6))
        self.rows_num_label.set_text(num_string_with_leading_zeros(self.full_rows_created, 6))
        
        # Render the five labels.   
        for label in self.labels:
            label.blit_to(surface)
        
        # Render the area showing the next tetris piece.
        self.render_next_piece_area(self.get_next_piece_area_surface(surface))
                    
    def render_block(self, surface, position, color):
        """
        Render a tetromino block on the given surface
        at the given tetris-board position with
        the given color.
        
        @param surface: a surface
        @type surface: C{pygame.Surface}
        
        @param position: (x,y) coordinate of the
            tetromino block in the board's coordinate
            system.
        @type position: 2-tuple of integers
        
        @param color: RGB color of the tetromino block.
        @type color: 3-tuple of integers
        """
        block_x, block_y = position
        if block_x >= 0 and block_x < self.board_width and block_y >= 0 and block_y < self.board_height:                    
            rect = Rect(self.block_size * block_x, 
                        self.block_size * block_y,
                        self.block_size,
                        self.block_size)
            pygame.draw.rect(surface, color, rect)
            pygame.draw.rect(surface, (0,0,0), rect, 1)
        
    def render_board(self, surface):
        """
        Render the tetris board and the tetromino
        the player is controlling to the
        given surface so that the 
        top left corner of the board
        is at coordinate (0,0) in 
        the given surface's coordinate system.
        """
        # Make the background black.
        surface.fill((0,0,0))
        
        # Render the board.
        for block in self.board.blocks:
            self.render_block(surface, block.position, tetromino_colors[block.shape])
        
        # Render the player's piece block by block.
        if self.player_piece is not None:
            for position in self.player_piece.block_positions():
                self.render_block(surface, position, tetromino_colors[self.player_piece.shape])
    
    def render_next_piece_area(self, surface):
        """
        Render the area showing the next tetromino
        to the given surface so that the 
        top left corner of the board
        is at coordinate (0,0) in 
        the given surface's coordinate system.
        """
        surface.fill((0,0,0))        
        positions = fixed_tetrominoes[self.next_piece.shape][0]
        for position in positions:
            position = (position[0]+2, position[1]+1)
            self.render_block(surface, position, tetromino_colors[self.next_piece.shape])
    
    def prepare_next_piece(self):
        """
        Pick a random tetromino piece to be the 
        next tetromino piece the player is going 
        to control.
        """
        # Promote the current next piece to
        # the piece the player controls.
        self.player_piece = self.next_piece
        
        # Pick a random shape.
        shape = fixed_tetrominoes.keys()[random.randint(0, len(fixed_tetrominoes)-1)]
        
        # Put the piece at the horizontal center
        # of the top row of the board.
        if shape != 'i':
            y_offset = -1
        else:
            y_offset = 0
        self.next_piece = Tetromino(shape, (self.board_width/2-1, y_offset))
            
    def switched_into(self, from_mode):
        """
        If the mode got switched into from 
        the Full Row Mode, then remove the 
        full rows and compute the score;
        otherwise, do nothing.
        """
        if from_mode == self.engine.full_row_mode:
            full_row_count = len(self.board.full_rows())
            self.full_rows_created += full_row_count
            self.score += self.full_row_score(full_row_count)
            self.board.remove_full_rows()
    
    def full_row_score(self, full_row_count):
        """
        Compute the score the player gets
        when creating the given number of 
        full rows.
        
        @param full_row_count: the number of 
            full rows
        @type full_row_count: integer
        
        @rtype: integer
        """
        return full_row_count**2 * self.board_width * self.piece_score()
    
    def piece_score(self):
        """
        Return the score the player gets
        when he drops a tetromino piece.
        """
        return 10
    
    def move_if_ok(self, new_position):
        """
        Try moving the player's tetromino so that
        the pivot block's position is at 
        the given new position, but cancel the
        move if the teromino collides with anything.
        
        @param new_position: (x,y) coordinate
            of the new position.
        @type new_position: 2-tuple of integers
        
        @return: True if the tetromino does not 
            hit anything after the move; otherwise, False.
        @rtype: boolean
        """
        old_position = self.player_piece.position
        self.player_piece.position = new_position
        if self.board.collide_with(self.player_piece):
            self.player_piece.position = old_position
            return False
        return True
    
    def rotate_if_ok(self):
        """
        Try rotating the player's tetromino counterclockwise, 
        but cancel the rotation if the teromino collides 
        with anything.
        
        @return: True if the tetromino does not 
            hit anything after the rotation; otherwise, False.
        @rtype: boolean
        """
        self.player_piece.rotate_counterclockwise()
        if self.board.collide_with(self.player_piece):
            self.player_piece.rotate_clockwise()
            return False
        return True
                        
    def handle(self, event):
        self.engine.handle(event)
        self.handle_key_events(event)   
                                
    def preprocess(self):
        self.clear_key_flags()
    
    def process(self):
        self.process_key_events()
        
        # Rotate if the user presses the up arrow.
        if self.is_key_down(pygame.locals.K_UP):
            self.rotate_if_ok()
        
        # Move left if the user types the left arrow.
        if self.is_key_typed(pygame.locals.K_LEFT):
            old_position = self.player_piece.position
            new_position = (old_position[0]-1, old_position[1])
            self.move_if_ok(new_position)
            
        # Move left if the user types the right arrow.
        if self.is_key_typed(pygame.locals.K_RIGHT):            
            old_position = self.player_piece.position
            new_position = (old_position[0]+1, old_position[1])
            self.move_if_ok(new_position)
                    
        # Moving down if
        # (1) it has been more than descend_delay 
        #     milliseconds since the last time the 
        #     player piece moved down, or 
        # (2) the user types the down arrow key.
        if (self.is_key_typed(pygame.locals.K_DOWN) or 
            self.millisecs_since_mark("last descent") > self.descend_delay):
            self.mark_time("last descent")
            old_x, old_y = self.player_piece.position
            new_position = (old_x, old_y+1)
            
            # If the move results in the tetromino collding
            # with something, then freeze the tetromino,
            # and let the player control the next tetromino piece.
            if not self.move_if_ok(new_position):
                self.board.freeze_tetromino(self.player_piece)
                self.prepare_next_piece()
                self.score += self.piece_score()
                
                # Also, we check if the tetromino makes
                # the row full. If so, we switch to
                # Ful Row mode to display the full rows.
                full_rows = self.board.full_rows()
                if len(full_rows) > 0:
                    self.engine.full_row_mode.full_rows = full_rows
                    self.engine.change_mode(self.engine.full_row_mode)
                    return
                
                # If the new tetromino collides with 
                # anything before it gets any chance
                # to move, then it's game over.
                if self.board.collide_with(self.player_piece):
                    self.engine.change_mode(self.engine.game_over_mode)
                    return