"""
Contain classes and variables that represent 
tetrominoes and the tetris board.

Representing Tetrominoes
========================
    Each Tetromino has a "pivot" block, which is 
    the block around with the tetromino is rotated.
    For example, the L tetromino has the middle block of 
    the three-block row as the pivot::

          *    *            **
        *X*    X    *X*      X
               **   *        *
               
    We establish a "local coordinate system" for each
    tetromino in such a way that the pivot is at position
    (0,0). Other blocks' coordinates are defined relative
    to the pivot. For example, the blocks of the tetromino::
    
        *
        X
        **
        
    have coordiates (0,-1), (0,0), (0,1), and (1,1), from
    top to bottom and left to right.
    
    As a result, the four rotations of the tetromino L can be represented
    by four lists of coordinates, each having four 2-tuples of 
    integers. Similarly, the tetromino I has only two lists as it
    has only two rotations. The tetromino O has only one list.
    
    For each type of tetromino, we put the lists representing their
    rotations together in a big list so that:
        1. The first list in the big list represents the "default" 
        rotation. This is the rotation that a new piece of tetromino
        is in when it shows up in the tetris board.
        2. The next list in the tetromino represents the tetromino
        rotated 90 degrees counterclockwise.
    In this way, a rotation can be specified by an integer index to 
    an element in the list. To rotate the tetromino counterclockwise, 
    we simply increase the index. Conversely, to rotate clockwise, 
    we decrease the index.
    
    To simplify the representation even further, we 
    create a dictionary named C{fixed_tetrominoes}, which maps shape
    characters (i, j, l, o, s, t, z) to the list of rotations of the
    tetromino of that shape. Hence, a tetromino in its own local coordinate
    can be represented by a character (which represents its shape),
    and an integer (which represents its rotation).
    
    To represent a tetromino floating above a tetris board, we keep
    the (x,y) position of the pivot block in the board's coordinate system.
    So, the positions of other blocks of the tetromino in the board's 
    coordinate system can be obtained by adding (x,y) to each 2-tuple in
    the list C{fixed_tetrominoes[shape][rotation]}, where C{shape} is 
    the character representing the shape, and C{rotation} is the integer 
    representing the rotation.
    
@var fixed_tetrominoes: A dictionary mapping shape characters --- i, j, l,
    o, s, t, and z --- to lists of rotations of the tetromino specified 
    by the shape character. Each rotation is represented by a list of (x,y)
    coordinates such that the pivot block has coordinate (0,0).
"""
fixed_tetrominoes = {
                     'o': [[(0,0),(1,0),(1,1),(0,1)]],
                     'i': [[(0,0),(-1,0),(1,0),(2,0)], [(0,0),(0,-1),(0,1),(0,2)]],
                     's': [[(0,0),(0,1),(-1,1),(1,0)], [(0,0),(0,-1),(1,0),(1,1)]],
                     'z': [[(0,0),(-1,0),(0,1),(1,1)], [(0,0),(0,1),(1,0),(1,-1)]],
                     'j': [[(0,0),(-1,0),(1,0),(1,1)], [(0,0),(0,-1),(0,1),(1,-1)],
                           [(0,0),(-1,0),(1,0),(-1,-1)], [(0,0),(0,-1),(0,1),(-1,1)]],
                     'l': [[(0,0),(-1,0),(1,0),(-1,1)], [(0,0),(0,-1),(0,1),(1,1)],
                           [(0,0),(-1,0),(1,0),(1,-1)], [(0,0),(0,-1),(0,1),(-1,-1)]],
                     't': [[(0,0),(-1,0),(1,0),(0,1)], [(0,0),(0,-1),(0,1),(1,0)],
                           [(0,0),(-1,0),(1,0),(0,-1)], [(0,0),(0,-1),(0,1),(-1,0)]]
                     }

class Tetromino:
    """
    A full piece of tetromino floating in the air.
    
    @ivar shape: the shape of the tetromino: i, j, l, o, s, t, or z.
    @type shape: character
    
    @ivar rotation: index to a particular rotation of the tetromino.
    @type rotation: integer
    
    @ivar position: M{(x,y)} coordinate of the pivot block of the tetromino
        in the coordinate system of the tetris board.
    @type position: 2-tuple of integers
    """
    def __init__(self, shape, position, rotation=0):
        """
        Create a floating tetromino piece with
        the given shape, position, and rotation.
        
        @param shape: the shape of the tetromino.
        @type shape: character
        @param rotation: the rotation of the tetromino
        @type rotation: integer
        @param position: the position in the board's coordinate system
            of the tetromino
        @type position: 2-tuple of integers
        """        
        self.shape = shape
        self.rotation = rotation
        self.position = position
    
    def rotate_counterclockwise(self):
        """
        Rotate the tetromino counterclockwise. This is done basically by
        cyclically incrementing C{rotation}.
        """
        self.rotation = (self.rotation + 1) % len(fixed_tetrominoes[self.shape])
        
    def rotate_clockwise(self):
        """
        Rotate the tetromino clockwise. This is done basically by
        cyclically decrementing C{rotation}.
        """
        self.rotation = (self.rotation - 1) % len(fixed_tetrominoes[self.shape])
            
    def block_positions(self):
        """
        Return the list of positions, in the board's
        coordinate system, of the blocks of this
        tetromino        
        
        @rtype: list of 2-tuples of integers
        """
        def displace(point):
            return (point[0] + self.position[0], point[1] + self.position[1])
        return map(displace, fixed_tetrominoes[self.shape][self.rotation])

class Block:
    """
    A fallen block of tetromino.
    
    @ivar position: M{(x,y)} coordinate of the block.
    @type position: 2-tuple of integers
    
    @ivar shape: the shape of the tetromino from which
        the block comes from.
    @type shape: character
    """
    def __init__(self, position, shape):
        """
        Create a tetromino block locating at the 
        specified position and coming from the
        tetromino with the specified shape.
        
        @param position: M{(x,y)} coordinate of the block.
        @type position: 2-tuple of integers
        @param shape: the shape of the tetromino from which
            the block comes from.
        @type shape: character
        """
        self.position = position
        self.shape = shape

class TetrisBoard:
    """
    A tetris board with tetromino blocks
    that have already fallen down.
    
    The board is a table with C{width} columns and C{height}
    rows. The columns are numbered 0 to C{width}-1 from
    left to right. The rows are numbered 0 to C{height}-1
    from top to bottom.
    
    @ivar width: the number of columns.
    @type width: integer
    
    @ivar height: the number of rows.
    @type height: integer
    
    @ivar blocks: the list of fallen tetromino blocks.
    @type blocks: list of C{Block}
    """
    def __init__(self, width, height):
        """
        Create an empty tetris board with C{width} columns
        and C{height} rows.        
        """
        self.width = width
        self.height = height
        self.blocks = []
        
    def sides_collide_with(self, tetromino):
        """
        Check if the given piece of tetromino
        has any block that lies outside the sides
        of the board. 
        
        @rtype: boolean
        """
        for x,y in tetromino.block_positions():
            if x < 0 or x >= self.width:
                return True
        return False
    
    def blocks_collide_with(self, tetromino):
        """
        Check if the given piece of tetromino
        has any block that shares position with any block
        in the board.
        
        @rtype: boolean
        """
        for x,y in tetromino.block_positions():
            for block in self.blocks:
                block_x, block_y = block.position
                if x == block_x and y == block_y:
                    return True                    
        return False
    
    def bottom_collide_with(self, tetromino):
        """
        Check if the given piece of tetromino
        has any block that lies below
        the bottom of the board.
        
        @rtype: boolean
        """
        for x,y in tetromino.block_positions():
            if y >= self.height:
                return True
        return False
    
    def collide_with(self, tetromino):
        """
        Check if the given piece of tetromino
        "collides" with either the bottom, any of the
        sides, or any of the blocks in the board.
        
        @rtype: boolean
        """
        return (self.sides_collide_with(tetromino) or 
                self.blocks_collide_with(tetromino) or 
                self.bottom_collide_with(tetromino))
                
    def freeze_tetromino(self, tetromino):
        """
        Change the given piece of tetromino to
        a number of blocks and store them in
        the board's list of blocks.
        """
        for position in tetromino.block_positions():
            self.blocks.append(Block(position, tetromino.shape))
            
    def full_rows(self):
        """
        Return a list of indices of rows in the board
        that are full.
        
        @rtype: list of integers
        """
        
        # Create a self.width x self.height array
        # whose elements are all False.
        grid = [[False for j in range(self.width)] for i in range(self.height)]
        # The above line is equal to the following lines:
        # grid = []
        # for i in range(self.height):
        #     row = []
        #     for j in range(self.widht):
        #         row.append(False)
        #     grid.append(row)
        
        # Fill the array "grid" so that
        # grid[y][x] = True if there is a block
        # at position (x,y).
        for block in self.blocks:
            x, y = block.position
            if x >= 0 and x < self.width and y >= 0 and y < self.height:
                grid[y][x] = True
        
        result = []
        # Loop through the rows, one by one.        
        for row in range(self.height):
            full = True
            # Check if the row is full. 
            for col in range(self.width):
                if not grid[row][col]:
                    full = False
                    break
            # If the row is full, then appends it to result.
            if full:
                result.append(row)
                                
        return result
    
    def remove_row(self, row_index):
        """
        Remove all blocks in the row specified by
        C{row_index}, and move all blocks above it down
        """
        
        def not_in_row(block):
            """
            Return True if and only if the
            given block's Y coordinate is not
            equal to row_index. 
            """
            return block.position[1] != row_index
        # Remove all blocks in the row specified by
        # row_index from self.blocks
        self.blocks = filter(not_in_row, self.blocks)
                
        def increase_y_if_above(block):
            """
            Return a new block which is identical
            to the given block if the block is below
            the row specified by row_index.
            
            However, if the block is above the row
            specified row_index, return the same block
            with its Y coordinate increased by 1.
            """
            if block.position[1] < row_index:
                return Block((block.position[0], block.position[1]+1), block.shape)
            else:
                return block
        # Increase the Y coordinate of all blocks above
        # the row specified by row_index.
        self.blocks = map(increase_y_if_above, self.blocks)
        
    def remove_full_rows(self):
        """
        Remove all full rows from the board.
        """        
        # Find the full rows.
        full_rows = self.full_rows()        
        # Then sort the row from top to bottom.
        full_rows.sort()        
        # And remove them one by one in that order.
        for row in full_rows:
            self.remove_row(row)
            
    def clear(self):
        """
        Remove all blocks from the board.
        """
        self.blocks = []