"""
Contain the mixin class TimeMarker.
"""

import pygame

class TimeMarker:
    """
    Mixin class that adds the capability to mark a certain
    moment in time and figure out how many milliseconds has
    elapsed since the marked time.

    Usage: The class that wants to use TimeMarker's capability
    first subclasses it, and call its constructor. Then, suppose 
    an event occurs just now and you want to mark it, 
    use the method::

        self.mark_time("name-of-the-event")

    where C{"name-of-the-event"} is a string or any hashable
    value that you use to denote the event. C{"name-of-the-event"}
    would be associated with the output of C{pygame.time.get_ticks()}
    After this, if you want to find out how many milliseconds 
    have elapsed since you marked the event, use the method::

        self.millisecs_since_marked("name-of-the-event")
    
    which returns the desired number of milliseconds.
    
    @ivar __marked_times: a dictionary mapping "names" of 
        events to the time it was marked.
    @type __marked_times: dictionary
    """ 
    def __init__(self):
        """
        Initialize the time marker.
        """   
        self.__marked_times = {}
    
    def mark_time(self, event_name):
        """
        Associate C{event_name} with the output
        of C{pygame.time.get_ticks()}.
        
        @param event_name: the value is that used
            to denote the event the user wants
            to mark.
        @type event_name: hashable
        """
        self.__marked_times[event_name] = pygame.time.get_ticks()
    
    def millisecs_since_mark(self, event_name):
        """
        Return the number of milliseconds since
        C{mark_time} was called with C{event_name}
        as the argument. If C{mark_time} with has never 
        been called with C{event_name}, return 0.
        
        @param event_name: the value is that used
            to denote an event the user marked.
        @type event_name: hashable
        """
        if event_name in self.__marked_times:
            return pygame.time.get_ticks() - self.__marked_times[event_name]
        else:
            return 0